import ipaddress
import re
from typing import Dict, Any, Union

from .parse import BaseParser, IPBaseKeys
from .servers import IPv4Allocations


class RIRParser(BaseParser):
    date_keys = (
        IPBaseKeys.ORG_REG_DATE,
        IPBaseKeys.ORG_UPDATED,
        IPBaseKeys.REG_DATE,
        IPBaseKeys.UPDATED,
    )

    # ARIN is used as "base" expression list
    reg_expressions = {
        IPBaseKeys.NET_RANGE: r"NetRange: *(.+)",
        IPBaseKeys.CIDR: r"CIDR: *(.+)",
        IPBaseKeys.NET_NAME: r"NetName: *(.+)",
        IPBaseKeys.NET_HANDLE: r"NetHandle: *(.+)",
        IPBaseKeys.NET_TYPE: r"NetType: *(.+)",
        IPBaseKeys.PARENT: r"Parent: *(.+)",
        IPBaseKeys.ORIGIN_AS: r"OriginAS: *(.+)",
        IPBaseKeys.ORGANIZATION: r"Organization: *(.+)",
        IPBaseKeys.REG_DATE: r"RegDate: *(.+)",
        IPBaseKeys.UPDATED: r"Updated: *(.+)",
        IPBaseKeys.RDAP_IP_REF: r"Updated: *.+\nRef: *(.+)",
        IPBaseKeys.ORG_NAME: r"OrgName: *(.+)",
        IPBaseKeys.ORG_ID: r"OrgId: *(.+)",
        IPBaseKeys.ORG_ADDRESS: r"Address: *(.+)",
        IPBaseKeys.ORG_CITY: r"City: *(.+)",
        IPBaseKeys.ORG_STATE: r"StateProv: *(.+)",
        IPBaseKeys.ORG_COUNTRY: r"Country: *(.+)",
        IPBaseKeys.ORG_ZIPCODE: r"PostalCode: *(.+)",
        IPBaseKeys.ORG_REG_DATE: r"Country: *.+\nRegDate: *(.+)",
        IPBaseKeys.ORG_UPDATED: r"Country: *.+\nRegDate: *.+\nUpdated: *(.+)",
        IPBaseKeys.ORG_RDAP_REF: r"Country: *.+\nRegDate: *.+\nUpdated: *.+\nRef: *(.+)",
        IPBaseKeys.ABUSE_HANDLE: r"OrgAbuseHandle: *(.+)",
        IPBaseKeys.ABUSE_NAME: r"OrgAbuseName: *(.+)",
        IPBaseKeys.ABUSE_PHONE: r"OrgAbusePhone: *(.+)",
        IPBaseKeys.ABUSE_EMAIL: r"OrgAbuseEmail: *(.+)",
        IPBaseKeys.ABUSE_ADDRESS: r"",  # not for ARIN, but for other RIRs
        IPBaseKeys.ABUSE_RDAP_REF: r"OrgAbuseRef: *(.+)",
        IPBaseKeys.ROUTING_HANDLE: r"OrgRoutingHandle: *(.+)",
        IPBaseKeys.ROUTING_NAME: r"OrgRoutingName: *(.+)",
        IPBaseKeys.ROUTING_PHONE: r"OrgRoutingPhone: *(.+)",
        IPBaseKeys.ROUTING_EMAIL: r"OrgRoutingEmail: *(.+)",
        IPBaseKeys.ROUTING_ADDRESS: r"",  # not for ARIN, but for other RIRs
        IPBaseKeys.ROUTING_RDAP_REF: r"OrgRoutingRef: *(.+)",
        IPBaseKeys.TECH_HANDLE: r"OrgTechHandle: *(.+)",
        IPBaseKeys.TECH_NAME: r"OrgTechName: *(.+)",
        IPBaseKeys.TECH_ADDRESS: r"",  # not for ARIN, but for other RIRs
        IPBaseKeys.TECH_PHONE: r"OrgTechPhone: *(.+)",
        IPBaseKeys.TECH_EMAIL: r"OrgTechEmail: *(.+)",
    }


class NumberParser:
    def __init__(self):
        self.servers = IPv4Allocations()

    def parse(
        self,
        blob: str,
        ip: Union[ipaddress.IPv4Address],
    ) -> dict[IPBaseKeys, Any]:
        _, server = self.servers.get_servers(ip)
        parser = self._init_parser(server)
        return parser.parse(blob)

    @staticmethod
    def _init_parser(rir_server: str) -> RIRParser:
        if rir_server == "whois.arin.net":
            return ARINParser()
        elif rir_server == "whois.afrinic.net":
            return AFRINICParser()
        elif rir_server == "whois.apnic.net":
            return APNICParser()
        elif rir_server == "whois.lacnic.net":
            return LACNICParser()
        elif rir_server == "whois.ripe.net":
            return RIPEParser()
        else:
            return ARINParser()  # default


class ARINParser(RIRParser):  # default
    def __init__(self): ...


class AFRINICParser(RIRParser):
    _afrinic = {
        IPBaseKeys.NET_RANGE: r"inetnum: *(.+)",
        IPBaseKeys.NET_NAME: r"netname: *(.+)",
        IPBaseKeys.NET_TYPE: r"status: *(.+)",
        IPBaseKeys.ORIGIN_AS: r"origin: *(.+)",
        IPBaseKeys.PARENT: r"parent: *(.+)",
        IPBaseKeys.ORGANIZATION: r"descr: *(.+)",
        IPBaseKeys.UPDATED: r"last-modified: *(.+)",
        IPBaseKeys.ORG_NAME: r"descr: *(.+)",
        IPBaseKeys.ORG_COUNTRY: r"country: *(.+)",
        IPBaseKeys.ABUSE_HANDLE: r"abuse-c: *(.+)",
        IPBaseKeys.ABUSE_NAME: r"",
        IPBaseKeys.ABUSE_ADDRESS: r"",
        IPBaseKeys.ABUSE_PHONE: r"",
        IPBaseKeys.ABUSE_EMAIL: r"",
        IPBaseKeys.ROUTING_HANDLE: r"admin-c: *(.+)",
        IPBaseKeys.ROUTING_NAME: r"",
        IPBaseKeys.ROUTING_ADDRESS: r"",
        IPBaseKeys.ROUTING_PHONE: r"",
        IPBaseKeys.ROUTING_EMAIL: r"",
        IPBaseKeys.TECH_HANDLE: r"tech-c: *(.+)",
        IPBaseKeys.TECH_NAME: r"",
        IPBaseKeys.TECH_ADDRESS: r"",
        IPBaseKeys.TECH_PHONE: r"",
        IPBaseKeys.TECH_EMAIL: r"",
    }

    _contact_fields = {
        "phone": r"phone: *(.+)",
        "name": r"(?:role|person): *(.+)",
        "address": r"address: *(.+)",
    }

    def __init__(self):
        self.update_reg_expressions(self._afrinic)

    def parse(self, blob: str) -> Dict[IPBaseKeys, Any]:
        parser_output = super().parse(blob)
        contact_field_fills = (
            ("abuse", parser_output.get(IPBaseKeys.ABUSE_HANDLE)),
            ("tech", parser_output.get(IPBaseKeys.TECH_HANDLE)),
            ("routing", parser_output.get(IPBaseKeys.ROUTING_HANDLE)),
        )
        # parse the contact info by looking up each "nic-hdl"
        for prefix, handle in contact_field_fills:
            if handle:
                pattern = re.compile(
                    r"(?:role|person):.+\n(?:.+\n){{1,}}nic-hdl: *{nic_hdl}\n(?:.+\n){{1,}}".format(
                        nic_hdl=handle
                    ),
                    flags=re.I,
                )
                contact_blob = pattern.search(blob)
                if contact_blob:
                    for field, field_regex in self._contact_fields.items():
                        key = getattr(IPBaseKeys, f"{prefix}_{field}".upper())
                        parser_output[key] = self.find_match(
                            field_regex, contact_blob.group()
                        )
        return parser_output


class APNICParser(RIRParser):
    _apnic = {
        IPBaseKeys.NET_RANGE: r"inetnum: *(.+)",
        IPBaseKeys.NET_NAME: r"netname: *(.+)",
        IPBaseKeys.NET_TYPE: r"status: *(.+)",
        IPBaseKeys.ORGANIZATION: r"descr: *(.+)",
        IPBaseKeys.UPDATED: r"last-modified: *(.+)",
        IPBaseKeys.ORG_NAME: r"irt: *(.+)",
        IPBaseKeys.ORG_ADDRESS: r"address: *(.+)",
        IPBaseKeys.ORG_COUNTRY: r"country: *(.+)",
        IPBaseKeys.ORG_UPDATED: r"irt:.+\n(?:.+\n){1,}last-modified: *(.+)",
        IPBaseKeys.ABUSE_HANDLE: r"abuse-c: *(.+)",
        IPBaseKeys.ABUSE_NAME: r"",
        IPBaseKeys.ABUSE_PHONE: r"",
        IPBaseKeys.ABUSE_EMAIL: r"",
        IPBaseKeys.ROUTING_HANDLE: r"admin-c: *(.+)",
        IPBaseKeys.ROUTING_NAME: r"",
        IPBaseKeys.ROUTING_PHONE: r"",
        IPBaseKeys.ROUTING_EMAIL: r"",
        IPBaseKeys.TECH_HANDLE: r"tech-c: *(.+)",
        IPBaseKeys.TECH_NAME: r"",
        IPBaseKeys.TECH_PHONE: r"",
        IPBaseKeys.TECH_EMAIL: r"",
    }

    _contact_fields = {
        "email": r"e-mail: *(.+)",
        "phone": r"phone: *(.+)",
        "name": r"(?:role|person): *(.+)",
        "address": r"address: *(.+)",
    }

    def __init__(self):
        self.update_reg_expressions(self._apnic)

    def parse(self, blob: str) -> Dict[IPBaseKeys, Any]:
        parser_output = super().parse(blob)
        contact_field_fills = (
            ("abuse", parser_output.get(IPBaseKeys.ABUSE_HANDLE)),
            ("tech", parser_output.get(IPBaseKeys.TECH_HANDLE)),
            ("routing", parser_output.get(IPBaseKeys.ROUTING_HANDLE)),
        )
        # parse the contact info by looking up each "nic-hdl"
        for prefix, handle in contact_field_fills:
            if handle:
                pattern = re.compile(
                    r"(?:role|person):.+\n(?:.+\n){{0,}}nic-hdl: *{nic_hdl}\n(?:.+\n){{1,}}".format(
                        nic_hdl=handle
                    ),
                    flags=re.I,
                )
                contact_blob = pattern.search(blob)
                if contact_blob:
                    for field, field_regex in self._contact_fields.items():
                        key = getattr(IPBaseKeys, f"{prefix}_{field}".upper())
                        if field == "address":
                            addresses = self.find_match(
                                field_regex, contact_blob.group(), many=True
                            )
                            parser_output[key] = (
                                ", ".join(addresses) if addresses else ""
                            )
                        else:
                            parser_output[key] = self.find_match(
                                field_regex, contact_blob.group()
                            )

        return parser_output


class LACNICParser(RIRParser):
    _lacnic = {
        IPBaseKeys.NET_RANGE: r"inetnum: *(.+)",
        IPBaseKeys.CIDR: r"inetrev: *(.+)",
        IPBaseKeys.NET_HANDLE: r"ownerid: *(.+)",
        IPBaseKeys.NET_TYPE: r"status: *(.+)",
        IPBaseKeys.ORGANIZATION: r"owner: *(.+)",
        IPBaseKeys.REG_DATE: r"created: *(.+)",
        IPBaseKeys.UPDATED: r"changed: *(.+)",
        IPBaseKeys.ORG_NAME: r"owner: *(.+)",
        IPBaseKeys.ORG_ID: r"ownerid: *(.+)",
        IPBaseKeys.ORG_ADDRESS: r"owner:.+\n(?:.+\n){1,}address: *(.+)",
        IPBaseKeys.ORG_COUNTRY: r"owner:.+\n(?:.+\n){1,}country: *(.+)",
        IPBaseKeys.ORG_REG_DATE: r"created: *(.+)",
        IPBaseKeys.ORG_UPDATED: r"changed: *(.+)",
        IPBaseKeys.ABUSE_HANDLE: r"abuse-c: *(.+)",
        IPBaseKeys.ABUSE_NAME: r"",  # handled in `parse`
        IPBaseKeys.ABUSE_PHONE: r"",
        IPBaseKeys.ABUSE_EMAIL: r"",
        IPBaseKeys.ROUTING_HANDLE: r"owner-c: *(.+)",
        IPBaseKeys.ROUTING_NAME: r"",
        IPBaseKeys.ROUTING_PHONE: r"",
        IPBaseKeys.ROUTING_EMAIL: r"",
        IPBaseKeys.TECH_HANDLE: r"tech-c: *(.+)",
        IPBaseKeys.TECH_NAME: r"",
        IPBaseKeys.TECH_PHONE: r"",
        IPBaseKeys.TECH_EMAIL: r"",
    }

    _contact_fields = {
        "email": r"e-mail: *(.+)",
        "phone": r"phone: *(.+)",
        "name": r"(?:role|person): *(.+)",
        "address": r"address: *(.+)",
    }

    def __init__(self):
        self.update_reg_expressions(self._lacnic)

    def parse(self, blob: str) -> Dict[IPBaseKeys, Any]:
        parser_output = super().parse(blob)
        contact_field_fills = (
            ("abuse", parser_output.get(IPBaseKeys.ABUSE_HANDLE)),
            ("tech", parser_output.get(IPBaseKeys.TECH_HANDLE)),
            ("routing", parser_output.get(IPBaseKeys.ROUTING_HANDLE)),
        )
        # parse contact info using each "nic-hdl"
        for prefix, handle in contact_field_fills:
            pattern = re.compile(
                r"nic-hdl(?:.){{0,}}:*{nic_hdl}\n(?:.+\n){{1,}}".format(nic_hdl=handle),
                flags=re.I,
            )
            contact_blob = pattern.search(blob)
            if contact_blob:
                for field, field_regex in self._contact_fields.items():
                    key = getattr(IPBaseKeys, f"{prefix}_{field}".upper())
                    if field == "address":
                        addresses = self.find_match(
                            field_regex, contact_blob.group(), many=True
                        )
                        parser_output[key] = ", ".join(addresses) if addresses else ""
                    else:
                        parser_output[key] = self.find_match(
                            field_regex, contact_blob.group()
                        )

        return parser_output


class RIPEParser(RIRParser):
    _ripe = {
        IPBaseKeys.NET_RANGE: r"inetnum: *(.+)",
        IPBaseKeys.NET_NAME: r"netname: *(.+)",
        IPBaseKeys.NET_TYPE: r"status: *(.+)",
        IPBaseKeys.ORIGIN_AS: r"origin: *(.+)",
        IPBaseKeys.ORGANIZATION: r"descr: *(.+)",
        IPBaseKeys.REG_DATE: r"created: *(.+)",
        IPBaseKeys.UPDATED: r"last-modified: *(.+)",
        IPBaseKeys.ABUSE_HANDLE: r"abuse-c: *(.+)",
        IPBaseKeys.ABUSE_NAME: r"",
        IPBaseKeys.ABUSE_PHONE: r"",
        IPBaseKeys.ABUSE_EMAIL: r"",
        IPBaseKeys.ROUTING_HANDLE: r"admin-c: *(.+)",
        IPBaseKeys.ROUTING_NAME: r"",
        IPBaseKeys.ROUTING_PHONE: r"",
        IPBaseKeys.ROUTING_EMAIL: r"",
        IPBaseKeys.TECH_HANDLE: r"tech-c: *(.+)",
        IPBaseKeys.TECH_NAME: r"",
        IPBaseKeys.TECH_PHONE: r"",
        IPBaseKeys.TECH_EMAIL: r"",
    }

    _contact_fields = {
        "email": r"abuse-mailbox: *(.+)",
        "phone": r"phone: *(.+)",
        "name": r"(?:role|person): *(.+)",
        "address": r"address: *(.+)",
    }

    def __init__(self):
        self.update_reg_expressions(self._ripe)

    def parse(self, blob: str) -> Dict[IPBaseKeys, Any]:
        parser_output = super().parse(blob)
        contact_field_fills = (
            ("abuse", parser_output.get(IPBaseKeys.ABUSE_HANDLE)),
            ("tech", parser_output.get(IPBaseKeys.TECH_HANDLE)),
            ("routing", parser_output.get(IPBaseKeys.ROUTING_HANDLE)),
        )
        # parse the contact info by looking up each "nic-hdl"
        for prefix, handle in contact_field_fills:
            if handle:
                pattern = re.compile(
                    r"(?:role|person):.+\n(?:.+\n){{1,}}nic-hdl: *{nic_hdl}\n(?:.+\n){{1,}}".format(
                        nic_hdl=handle
                    ),
                    flags=re.I,
                )
                contact_blob = pattern.search(blob)
                if contact_blob:
                    for field, field_regex in self._contact_fields.items():
                        key = getattr(IPBaseKeys, f"{prefix}_{field}".upper())
                        if field == "address":
                            addresses = self.find_match(
                                field_regex, contact_blob.group(), many=True
                            )
                            parser_output[key] = (
                                ", ".join(addresses) if addresses else ""
                            )
                        else:
                            parser_output[key] = self.find_match(
                                field_regex, contact_blob.group()
                            )

        return parser_output



"""
This file has been generated by ipv4-allocations-update.py program.

If you need it updated, override this file content with ipv4-allocations-update.py output.
"""

from ipaddress import IPv4Network


AllocationsT = dict[IPv4Network, dict[str, str]]


IPV4_ALLOCATIONS: AllocationsT = {
    IPv4Network("1.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("14.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("27.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("36.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("39.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("42.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("49.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("58.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("60.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("101.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("103.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("106.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("110.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("112.0.0.0/5"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("120.0.0.0/6"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("124.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("126.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("133.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("150.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("153.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("163.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("171.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("175.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("180.0.0.0/8"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("182.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("202.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("210.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("218.0.0.0/7"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("220.0.0.0/6"): {
        "rdap": "https://rdap.apnic.net/",
        "whois": "whois.apnic.net",
    },
    IPv4Network("2.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("5.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("25.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("31.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("37.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("46.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("51.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("53.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("57.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("62.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("77.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("78.0.0.0/7"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("80.0.0.0/4"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("109.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("141.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("145.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("151.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("176.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("178.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("185.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("188.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("193.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("194.0.0.0/7"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("212.0.0.0/7"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("217.0.0.0/8"): {
        "rdap": "https://rdap.db.ripe.net/",
        "whois": "whois.ripe.net",
    },
    IPv4Network("3.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("4.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("6.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("8.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("11.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("12.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("15.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("16.0.0.0/5"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("24.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("26.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("28.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("30.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("32.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("38.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("40.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("44.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("47.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("48.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("50.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("52.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("54.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("56.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("63.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("64.0.0.0/5"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("72.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("76.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("96.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("100.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("104.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("107.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("108.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("128.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("132.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("134.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("136.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("140.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("142.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("144.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("146.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("148.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("152.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("155.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("156.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("160.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("162.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("164.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("168.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("170.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("172.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("174.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("184.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("192.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("198.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("204.0.0.0/6"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("208.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("214.0.0.0/7"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("216.0.0.0/8"): {
        "rdap": "https://rdap.arin.net/registry",
        "whois": "whois.arin.net",
    },
    IPv4Network("41.0.0.0/8"): {
        "rdap": "https://rdap.afrinic.net/rdap/",
        "whois": "whois.afrinic.net",
    },
    IPv4Network("102.0.0.0/8"): {
        "rdap": "https://rdap.afrinic.net/rdap/",
        "whois": "whois.afrinic.net",
    },
    IPv4Network("105.0.0.0/8"): {
        "rdap": "https://rdap.afrinic.net/rdap/",
        "whois": "whois.afrinic.net",
    },
    IPv4Network("154.0.0.0/8"): {
        "rdap": "https://rdap.afrinic.net/rdap/",
        "whois": "whois.afrinic.net",
    },
    IPv4Network("196.0.0.0/7"): {
        "rdap": "https://rdap.afrinic.net/rdap/",
        "whois": "whois.afrinic.net",
    },
    IPv4Network("177.0.0.0/8"): {
        "rdap": "https://rdap.lacnic.net/rdap/",
        "whois": "whois.lacnic.net",
    },
    IPv4Network("179.0.0.0/8"): {
        "rdap": "https://rdap.lacnic.net/rdap/",
        "whois": "whois.lacnic.net",
    },
    IPv4Network("181.0.0.0/8"): {
        "rdap": "https://rdap.lacnic.net/rdap/",
        "whois": "whois.lacnic.net",
    },
    IPv4Network("186.0.0.0/7"): {
        "rdap": "https://rdap.lacnic.net/rdap/",
        "whois": "whois.lacnic.net",
    },
    IPv4Network("189.0.0.0/8"): {
        "rdap": "https://rdap.lacnic.net/rdap/",
        "whois": "whois.lacnic.net",
    },
    IPv4Network("190.0.0.0/7"): {
        "rdap": "https://rdap.lacnic.net/rdap/",
        "whois": "whois.lacnic.net",
    },
    IPv4Network("200.0.0.0/7"): {
        "rdap": "https://rdap.lacnic.net/rdap/",
        "whois": "whois.lacnic.net",
    },
}


__all__ = ("IPV4_ALLOCATIONS", "AllocationsT")



import ipaddress
from ipaddress import IPv4Address, IPv6Address
from typing import Any, Optional, Union
from dataclasses import dataclass
from warnings import warn

import whodap
from tldextract.tldextract import TLDExtract

from .client import (
    ASNClient,
    DomainClient,
    NumberClient,
    convert_to_ip,
)
from .errors import NotFoundError, GeneralError, QueryError, WhoIsError

__all__ = [
    "aio_rdap",
    "aio_whois",
    "whois",
    "rdap",
    "aio_whois_domain",
    "aio_whois_ipv4",
    "aio_whois_ipv6",
    "aio_rdap_domain",
    "aio_rdap_ipv4",
    "aio_rdap_ipv6",
    "aio_rdap_asn",
    "rdap_domain",
    "rdap_ipv4",
    "rdap_ipv6",
    "rdap_asn",
    "whois_domain",
    "whois_ipv4",
    "whois_ipv6",
    "ASNClient",
    "DomainClient",
    "NumberClient",
    "NotFoundError",
    "WhoIsError",
    "GeneralError",
    "QueryError",
]
__version__ = "1.1.5"


def whois(
    search_term: Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address],
    authoritative_only: bool = False,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> tuple[str, dict]:
    """
    Performs a WHOIS query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the query output
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return NumberClient(
            authoritative_only=authoritative_only,
            proxy_url=proxy_url,
            timeout=timeout,
        ).whois(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return NumberClient(
                authoritative_only=authoritative_only,
                proxy_url=proxy_url,
                timeout=timeout,
            ).whois(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return DomainClient(
                authoritative_only=authoritative_only,
                ignore_not_found=ignore_not_found,
                proxy_url=proxy_url,
                timeout=timeout,
                tldextract_obj=tldextract_obj,
            ).whois(search_term)
    else:
        return "", {}


def rdap(
    search_term: Union[int, str, ipaddress.IPv4Address, ipaddress.IPv6Address],
    authoritative_only: bool = False,
    whodap_client: Union[
        whodap.DNSClient, whodap.IPv4Client, whodap.IPv6Client, whodap.ASNClient
    ] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> tuple[str, dict]:
    """
    Performs an RDAP query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param whodap_client: An optional preconfigured instance of either an ASN, DNS, IPv4, or IPv6 `whodap` client
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return NumberClient(
            authoritative_only=authoritative_only,
            whodap_client=whodap_client,
        ).rdap(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return NumberClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
            ).rdap(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return DomainClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
                tldextract_obj=tldextract_obj,
            ).rdap(search_term)
    elif isinstance(search_term, int):
        return ASNClient(whodap_client=whodap_client).rdap(search_term)
    else:
        return "", {}


async def aio_whois(
    search_term: str,
    authoritative_only: bool = False,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> tuple[str, dict]:
    """
    Performs a WHOIS query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the query output
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return await NumberClient(
            authoritative_only=authoritative_only,
            proxy_url=proxy_url,
            timeout=timeout,
        ).aio_whois(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return await NumberClient(
                authoritative_only=authoritative_only,
                proxy_url=proxy_url,
                timeout=timeout,
            ).aio_whois(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return await DomainClient(
                authoritative_only=authoritative_only,
                ignore_not_found=ignore_not_found,
                proxy_url=proxy_url,
                timeout=timeout,
                tldextract_obj=tldextract_obj,
            ).aio_whois(search_term)
    else:
        return "", {}


async def aio_rdap(
    search_term: Union[int, str, ipaddress.IPv4Address, ipaddress.IPv6Address],
    authoritative_only: bool = False,
    whodap_client: Union[
        whodap.DNSClient, whodap.IPv4Client, whodap.IPv6Client, whodap.ASNClient
    ] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> tuple[str, dict]:
    """
    Performs an RDAP query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param whodap_client: An optional preconfigured instance of either an ASN, DNS, IPv4, or IPv6 `whodap` client
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return await NumberClient(
            authoritative_only=authoritative_only,
            whodap_client=whodap_client,
        ).aio_rdap(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return await NumberClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
            ).aio_rdap(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return await DomainClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
                tldextract_obj=tldextract_obj,
            ).aio_rdap(search_term)
    elif isinstance(search_term, int):
        return await ASNClient(whodap_client=whodap_client).aio_rdap(search_term)
    else:
        return "", {}


# ====================
# TODO: ALL the code below will be removed in a future release; it is here for backwards compatibility only


@dataclass
class DomainLookup:
    query_output: str
    parser_output: dict


@dataclass
class NumberLookup:
    query_output: str
    parser_output: dict


@dataclass
class ASNLookup:
    query_output: str
    parser_output: dict


def whois_domain(
    domain: str,
    authoritative_only: bool = False,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> DomainLookup:
    """
    Performs domain lookups with WHOIS.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the `query_output`
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.whois_domain` is deprecated. Please use `asyncwhois.whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = DomainClient(
        authoritative_only=authoritative_only,
        ignore_not_found=ignore_not_found,
        proxy_url=proxy_url,
        timeout=timeout,
        tldextract_obj=tldextract_obj,
    ).whois(domain)
    return DomainLookup(query_output, parser_output)


async def aio_whois_domain(
    domain: str,
    authoritative_only: bool = False,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> DomainLookup:
    """
    Performs asynchronous domain lookups with WHOIS.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the `query_output`
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.aio_whois_domain` is deprecated. Please use `asyncwhois.aio_whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = await DomainClient(
        authoritative_only=authoritative_only,
        ignore_not_found=ignore_not_found,
        proxy_url=proxy_url,
        timeout=timeout,
        tldextract_obj=tldextract_obj,
    ).aio_whois(domain)
    return DomainLookup(query_output, parser_output)


def rdap_domain(
    domain: str,
    httpx_client: Optional[Any] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> DomainLookup:
    """
    Performs an RDAP query for the given domain.
    Finds the authoritative RDAP server and parses the response from that server.

    :param domain: Any domain name or URL
        (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param httpx_client: Optional preconfigured instance of `httpx.AsyncClient`
    :param tldextract_obj: Optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.rdap_domain` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.DNSClient.new_client(httpx_client=httpx_client)
    query_output, parser_output = DomainClient(
        whodap_client=whodap_client,
        tldextract_obj=tldextract_obj,
    ).rdap(domain)
    return DomainLookup(query_output, parser_output)


async def aio_rdap_domain(
    domain: str,
    httpx_client: Optional[Any] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> DomainLookup:
    """
    Performs an async RDAP query for the given domain name.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param httpx_client: Optional preconfigured instance of `httpx.AsyncClient`
    :param tldextract_obj: Optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.aio_rdap_domain` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.DNSClient.new_aio_client(httpx_client=httpx_client)
    query_output, parser_output = await DomainClient(
        whodap_client=whodap_client,
        tldextract_obj=tldextract_obj,
    ).aio_rdap(domain)
    return DomainLookup(query_output, parser_output)


def whois_ipv4(
    ipv4: Union[IPv4Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs a WHOIS query for the given IPv4 address.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param ipv4: ip address as a str or `ipaddress.IPv4Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.whois_ipv4` is deprecated. Please use `asyncwhois.whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).whois(ipv4)
    return NumberLookup(query_output, parser_output)


async def aio_whois_ipv4(
    ipv4: Union[IPv4Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs an async WHOIS query for the given IPv4 address.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param ipv4: ip address as a str or `ipaddress.IPv4Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_whois_ipv4` is deprecated. Please use `asyncwhois.aio_whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = await NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).aio_whois(ipv4)
    return NumberLookup(query_output, parser_output)


def rdap_ipv4(
    ipv4: Union[IPv4Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an RDAP query for the given IPv4 address.

    :param ipv4: IP address as a string or `ipaddress.IPv4Address` object
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.rdap_ipv4` is deprecated. Please use `asyncwhois.rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv4Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).rdap(ipv4)
    return NumberLookup(query_output, parser_output)


async def aio_rdap_ipv4(
    ipv4: Union[IPv4Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an async RDAP query for the given IPv6 address.

    :param ipv4: IP address as a string or `ipaddress.IPv4Address` object
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_rdap_ipv4` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv4Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).aio_rdap(ipv4)
    return NumberLookup(query_output, parser_output)


def whois_ipv6(
    ipv6: Union[IPv6Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs a WHOIS query for the given IPv6 address.
    Looks up the WHOIS server, submits the query, and then parses the response from the server.

    :param ipv6: ip address as a str or `ipaddress.IPv6address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.whois_ipv6` is deprecated. Please use `asyncwhois.whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).whois(ipv6)
    return NumberLookup(query_output, parser_output)


async def aio_whois_ipv6(
    ipv6: Union[IPv6Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs an async WHOIS query for the given IPv6 address.
    Looks up the WHOIS server, submits the query, and then parses the response from the server.

    :param ipv6: ip address as a str or `ipaddress.IPv6Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_whois_ipv6` is deprecated. Please use `asyncwhois.aio_whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = await NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).aio_whois(ipv6)
    return NumberLookup(query_output, parser_output)


def rdap_ipv6(
    ipv6: Union[IPv6Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an RDAP query for the given IPv6 address.

    :param ipv6: IP address as a string or `ipaddress.IPv6Address` object
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.rdap_ipv6` is deprecated. Please use `asyncwhois.rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).rdap(ipv6)
    return NumberLookup(query_output, parser_output)


async def aio_rdap_ipv6(
    ipv6: Union[IPv6Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an async RDAP query for the given IPv6 address.

    :param ipv6: IP address as a string or `ipaddress.IPv6Address` object
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_rdap_ipv6` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_aio_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).aio_rdap(ipv6)
    return NumberLookup(query_output, parser_output)


def rdap_asn(asn: int, httpx_client: Optional[Any] = None) -> ASNLookup:
    """
    Performs an RDAP query for the given Autonomous System Number.

    :param asn: The ASN number as an integer
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of ASNLookup
    """
    warn(
        "`asyncwhois.rdap_asn` is deprecated. Please use `asyncwhois.rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = ASNClient(
        whodap_client=whodap_client,
    ).rdap(asn)
    return ASNLookup(query_output, parser_output)


async def aio_rdap_asn(asn: int, httpx_client: Optional[Any] = None) -> ASNLookup:
    """
    Performs an async RDAP query for the given Autonomous System Number.

    :param asn: The ASN number as an integer
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of ASNLookup
    """
    warn(
        "`asyncwhois.aio_rdap_asn` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_aio_client(httpx_client=httpx_client)
    query_output, parser_output = ASNClient(
        whodap_client=whodap_client,
    ).aio_rdap(asn)
    return ASNLookup(query_output, parser_output)



import asyncio
import time

import asyncwhois


async def main():
    urls = [
        'https://www.google.co.uk',
        'en.wikipedia.org/wiki/Pi',
        'https://www.urbandictionary.com/define.php?term=async',
        'twitch.tv',
        'https://www.pcpartpicker.com',
        'https://packaging.python.org/',
        'https://www.amazon.co.jp',
        'github.com/explore',
        'mango.beer'
    ]
    futures = []
    for url in urls:
        future = asyncwhois.whois_domain(url)
        futures.append(future)

    whois_data = await asyncio.gather(*futures)


if __name__ == '__main__':
    start = time.time()
    asyncio.run(main())
    print(f'Done! [{round(time.time() - start, 4)}] seconds.')



import ipaddress
from typing import Union, Any

from tldextract.tldextract import extract, TLDExtract
import whodap

from .parse import convert_whodap_keys, IPBaseKeys, TLDBaseKeys
from .parse_rir import NumberParser
from .parse_tld import DomainParser
from .query import DomainQuery, NumberQuery


def convert_to_ip(ip: str):
    try:
        ip_obj = ipaddress.ip_address(ip)
        return ip_obj
    except ipaddress.AddressValueError as e:
        raise e


class Client:
    def __init__(self, whodap_client):
        self.whodap_client = whodap_client

    def init_whodap_client(self, ipv4: bool = True):
        if isinstance(self, DomainClient):
            self.whodap_client = whodap.DNSClient.new_client()
        elif isinstance(self, NumberClient):
            if ipv4:
                self.whodap_client = whodap.IPv4Client.new_client()
            else:
                self.whodap_client = whodap.IPv6Client.new_client()
        elif isinstance(self, ASNClient):
            self.whodap_client = whodap.ASNClient.new_client()

    async def init_async_whodap_client(self, ipv4: bool = True):
        if isinstance(self, DomainClient):
            self.whodap_client = await whodap.DNSClient.new_aio_client()
        elif isinstance(self, NumberClient):
            if ipv4:
                self.whodap_client = await whodap.IPv4Client.new_aio_client()
            else:
                self.whodap_client = await whodap.IPv6Client.new_aio_client()
        elif isinstance(self, ASNClient):
            self.whodap_client = await whodap.ASNClient.new_aio_client()


class DomainClient(Client):
    def __init__(
        self,
        authoritative_only: bool = False,
        ignore_not_found: bool = False,
        proxy_url: str = None,
        whodap_client: whodap.DNSClient = None,
        timeout: int = 10,
        tldextract_obj: TLDExtract = None,
    ):
        super().__init__(whodap_client)
        self.authoritative_only = authoritative_only
        self.ignore_not_found = ignore_not_found
        self.proxy_url = proxy_url
        self.timeout = timeout
        self.tldextract_obj = tldextract_obj
        self.query_obj = DomainQuery(proxy_url=proxy_url, timeout=timeout)
        self.parse_obj = DomainParser(ignore_not_found=ignore_not_found)

    def _get_domain_components(self, domain: str) -> tuple[str, str, str]:
        ext = (
            extract(domain)
            if self.tldextract_obj is None
            else self.tldextract_obj(domain)
        )
        suffix = ext.suffix.split(".")[-1]
        return ext.registered_domain, ext.domain, suffix

    def rdap(self, domain: str) -> tuple[str, dict]:
        if self.whodap_client is None:
            self.init_whodap_client()
        _, domain, tld = self._get_domain_components(domain)
        rdap_output = self.whodap_client.lookup(domain, tld)
        query_string = rdap_output.to_json()
        parsed_dict = convert_whodap_keys(rdap_output.to_whois_dict())
        return query_string, parsed_dict

    def whois(self, domain: str) -> tuple[str, dict[TLDBaseKeys, Any]]:
        registered_domain, domain, tld = self._get_domain_components(domain)
        query_chain: list[str] = self.query_obj.run(registered_domain)
        authoritative_answer = query_chain[-1]
        parsed_dict: dict[TLDBaseKeys, Any] = self.parse_obj.parse(
            authoritative_answer, tld
        )
        query_string = (
            authoritative_answer if self.authoritative_only else "\n".join(query_chain)
        )
        return query_string, parsed_dict

    async def aio_rdap(self, domain: str) -> tuple[str, dict]:
        if self.whodap_client is None:
            await self.init_async_whodap_client()
        _, domain, tld = self._get_domain_components(domain)
        rdap_output = await self.whodap_client.aio_lookup(domain, tld)
        query_string = rdap_output.to_json()
        parsed_dict = convert_whodap_keys(rdap_output.to_whois_dict())
        return query_string, parsed_dict

    async def aio_whois(self, domain: str) -> tuple[str, dict[TLDBaseKeys, Any]]:
        registered_domain, domain, tld = self._get_domain_components(domain)
        query_chain: list[str] = await self.query_obj.aio_run(registered_domain)
        authoritative_answer = query_chain[-1]
        parsed_dict: dict[TLDBaseKeys, Any] = self.parse_obj.parse(
            authoritative_answer, tld
        )
        query_string = (
            authoritative_answer if self.authoritative_only else "\n".join(query_chain)
        )
        return query_string, parsed_dict


class NumberClient(Client):
    def __init__(
        self,
        authoritative_only: bool = False,
        proxy_url: str = None,
        whodap_client: Union[whodap.IPv4Client, whodap.IPv6Client] = None,
        timeout: int = 10,
    ):
        super().__init__(whodap_client)
        self.authoritative_only = authoritative_only
        self.proxy_url = proxy_url
        self.timeout = timeout
        self.whodap_client = whodap_client
        self.query_obj = NumberQuery(proxy_url=proxy_url, timeout=timeout)
        self.parse_obj = NumberParser()

    def rdap(
        self, ip: Union[ipaddress.IPv4Address, ipaddress.IPv6Address, str]
    ) -> tuple[str, dict]:
        if not isinstance(ip, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            ip = convert_to_ip(ip)
        if self.whodap_client is None:
            self.init_whodap_client(ipv4=(ip.version == 4))
        query_string = self.whodap_client.lookup(ip).to_json()
        return query_string, {}  # no parsed output available

    def whois(
        self, ip: Union[ipaddress.IPv4Address, ipaddress.IPv6Address, str]
    ) -> tuple[str, dict[IPBaseKeys, Any]]:
        if not isinstance(ip, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            ip = convert_to_ip(ip)
        query_chain: list[str] = self.query_obj.run(ip)
        authoritative_answer = query_chain[-1]
        parsed_dict: dict[IPBaseKeys, Any] = {}
        if isinstance(ip, ipaddress.IPv4Address):
            parsed_dict = self.parse_obj.parse(authoritative_answer, ip)
        query_string = (
            authoritative_answer if self.authoritative_only else "\n".join(query_chain)
        )
        return query_string, parsed_dict

    async def aio_rdap(
        self, ip: Union[ipaddress.IPv4Address, ipaddress.IPv6Address, str]
    ) -> tuple[str, dict]:
        if not isinstance(ip, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            ip = convert_to_ip(ip)
        if self.whodap_client is None:
            await self.init_async_whodap_client(ipv4=(ip.version == 4))
        query_resp = await self.whodap_client.aio_lookup(ip)
        query_string = query_resp.to_json()
        return query_string, {}  # no parsed output available

    async def aio_whois(
        self, ip: Union[ipaddress.IPv4Address, ipaddress.IPv6Address, str]
    ) -> tuple[str, dict[IPBaseKeys, Any]]:
        if not isinstance(ip, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            ip = convert_to_ip(ip)
        query_chain: list[str] = await self.query_obj.aio_run(ip)
        authoritative_answer = query_chain[-1]
        parsed_dict: dict[IPBaseKeys, Any] = {}
        if isinstance(ip, ipaddress.IPv4Address):
            parsed_dict = self.parse_obj.parse(authoritative_answer, ip)
        query_string = (
            authoritative_answer if self.authoritative_only else "\n".join(query_chain)
        )
        return query_string, parsed_dict


class ASNClient(Client):
    def __init__(
        self,
        whodap_client: whodap.ASNClient = None,
        timeout: int = 10,
    ):
        super().__init__(whodap_client)
        self.timeout = timeout

    def rdap(self, asn: int) -> tuple[str, dict]:
        if self.whodap_client is None:
            self.init_whodap_client()
        query_resp = self.whodap_client.lookup(asn)
        query_string = query_resp.to_json()
        return query_string, {}

    async def aio_rdap(self, asn: int) -> tuple[str, dict]:
        if self.whodap_client is None:
            await self.init_async_whodap_client()
        query_resp = await self.whodap_client.aio_lookup(asn)
        query_string = query_resp.to_json()
        return query_string, {}



import asyncio
import ipaddress
import re
import socket
from typing import Tuple, Generator, Union
from contextlib import contextmanager, asynccontextmanager

from python_socks.sync import Proxy
from python_socks.async_.asyncio import Proxy as AsyncProxy

from .servers import IPv4Allocations, CountryCodeTLD, GenericTLD, SponsoredTLD

BLOCKSIZE = 1500


class Query:
    iana_server = "whois.iana.org"
    whois_port = 43
    refer_regex = r"refer: *(.+)"
    whois_server_regex = r".+ whois server: *(.+)"

    def __init__(self, proxy_url: str = None, timeout: int = 10):
        self.proxy_url = proxy_url
        self.timeout = timeout

    @staticmethod
    def _find_match(regex: str, blob: str) -> str:
        match = ""
        found = re.search(regex, blob, flags=re.IGNORECASE)
        if found:
            match = found.group(1).rstrip("\r").replace(" ", "").rstrip(":").rstrip("/")
        return match

    @contextmanager
    def _create_connection(
        self, address: Tuple[str, int], proxy_url: str = None
    ) -> Generator[socket.socket, None, None]:
        s = None
        try:
            # Use proxy if specified
            if proxy_url:
                proxy = Proxy.from_url(proxy_url)
                # proxy is a standard python socket in blocking mode
                s = proxy.connect(*address, timeout=self.timeout)
            else:
                # otherwise use socket
                s = socket.create_connection(address, self.timeout)
            yield s
        finally:
            if s and hasattr(s, "close"):
                s.close()

    @asynccontextmanager
    async def _aio_create_connection(
        self, address: Tuple[str, int], proxy_url: str = None
    ) -> Generator[Tuple[asyncio.StreamReader, asyncio.StreamWriter], None, None]:
        # init
        reader, writer = None, None
        # Use proxy if specified
        if proxy_url:
            proxy = AsyncProxy.from_url(proxy_url)
            # sock is a standard python socket in blocking mode
            sock = await proxy.connect(*address, timeout=self.timeout)
            # pass it to asyncio
            s = asyncio.open_connection(host=None, port=None, sock=sock)
        else:
            # otherwise use asyncio to open the connection
            s = asyncio.open_connection(*address)
        try:
            reader, writer = await asyncio.wait_for(s, self.timeout)
            yield reader, writer
        finally:
            if writer:
                writer.close()
                if hasattr(writer, "wait_closed"):
                    await writer.wait_closed()

    @staticmethod
    def _send_and_recv(conn: socket.socket, data: str) -> str:
        conn.sendall(data.encode())
        result = ""
        while True:
            received = conn.recv(BLOCKSIZE)
            if received == b"":
                break
            else:
                result += received.decode("utf-8", errors="ignore")
        return result

    @staticmethod
    async def _aio_send_and_recv(
        reader: asyncio.StreamReader, writer: asyncio.StreamWriter, data: str
    ) -> str:
        writer.write(data.encode())
        result = ""
        while True:
            received = await reader.read(BLOCKSIZE)
            if received == b"":
                break
            else:
                result += received.decode("utf-8", errors="ignore")
        return result

    def run(self, search_term: str, server: str = None) -> list[str]:
        """
        Submits the `search_term` to the WHOIS server and returns a list of query responses.
        """
        data = search_term + "\r\n"
        if not server:
            # TODO: think about moving this to subclass
            if ":" in data:  # ipv6
                server_regex = r"whois: *(.+)"
            else:
                server_regex = self.refer_regex
            server = self.iana_server
        else:
            server_regex = self.whois_server_regex
        return self._do_query(server, data, server_regex, [])

    async def aio_run(self, search_term: str, server: str = None) -> list[str]:
        data = search_term + "\r\n"
        if not server:
            if ":" in data:  # ipv6
                server_regex = r"whois: *(.+)"
            else:
                server_regex = self.refer_regex
            server = self.iana_server
        else:
            server_regex = self.whois_server_regex
        return await self._aio_do_query(server, data, server_regex, [])

    def _do_query(
        self, server: str, data: str, regex: str, chain: list[str]
    ) -> list[str]:
        """
        Recursively submits WHOIS queries until it reaches the Authoritative Server.
        """
        # connect to whois://<server>:43
        with self._create_connection((server, self.whois_port), self.proxy_url) as conn:
            # submit domain and receive raw query output
            query_output = self._send_and_recv(conn, data)
            # save query chain
            chain.append(query_output)
            # parse response for the referred WHOIS server name
            whois_server = self._find_match(regex, query_output)
            whois_server = whois_server.lower()
            if (
                whois_server
                and whois_server != server
                and not whois_server.startswith("http")
            ):
                # recursive call to find more authoritative server
                chain = self._do_query(
                    whois_server, data, self.whois_server_regex, chain
                )
        # return the WHOIS query chain
        return chain

    async def _aio_do_query(
        self, server: str, data: str, regex: str, chain: list[str]
    ) -> list[str]:
        # connect to whois://<server>:43
        async with self._aio_create_connection(
            (server, self.whois_port), self.proxy_url
        ) as r_and_w:
            # socket reader and writer
            reader, writer = r_and_w
            # submit domain and receive raw query output
            query_output = await asyncio.wait_for(
                self._aio_send_and_recv(reader, writer, data), self.timeout
            )
            chain.append(query_output)
            # parse response for the referred WHOIS server name
            whois_server = self._find_match(regex, query_output)
            whois_server = whois_server.lower()
            if (
                whois_server
                and whois_server != server
                and not whois_server.startswith("http")
            ):
                # recursive call to find the authoritative server
                chain = await self._aio_do_query(
                    whois_server, data, self.whois_server_regex, chain
                )
        # return the WHOIS query chain
        return chain


class DomainQuery(Query):
    def __init__(
        self,
        server: str = None,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        super().__init__(proxy_url, timeout)
        self.server = server

    @staticmethod
    def _get_server_name(tld: str) -> Union[str, None]:
        tld_converted = tld.upper().replace("-", "_")
        for servers in [CountryCodeTLD, GenericTLD, SponsoredTLD]:
            if hasattr(servers, tld_converted):
                server = getattr(servers, tld_converted)
                return server
        return None

    def run(self, search_term: str, server: str = None) -> list[str]:
        if not server:
            server = self._get_server_name(search_term)
        return super().run(str(search_term), server)

    async def aio_run(self, search_term: str, server: str = None) -> list[str]:
        if not server:
            server = self._get_server_name(search_term)
        return await super().aio_run(str(search_term), server)


class NumberQuery(Query):
    def __init__(
        self,
        server: str = None,
        proxy_url: str = None,
        timeout: int = 10,
    ):
        super().__init__(proxy_url, timeout)
        self.server = server
        self.whois_server_regex = r"ReferralServer: *whois://(.+)"

    @staticmethod
    def _get_server_name(ip: Union[ipaddress.IPv4Address, ipaddress.IPv6Address]):
        if isinstance(ip, ipaddress.IPv4Address):
            _, server = IPv4Allocations().get_servers(ip)
            return server
        elif isinstance(ip, ipaddress.IPv6Address):
            return None

    def run(
        self,
        search_term: Union[ipaddress.IPv4Address, ipaddress.IPv6Address],
        server: str = None,
    ) -> list[str]:
        if not server:
            server = self._get_server_name(search_term)
        return super().run(str(search_term), server)

    async def aio_run(
        self,
        search_term: Union[ipaddress.IPv4Address, ipaddress.IPv6Address],
        server: str = None,
    ) -> list[str]:
        if not server:
            server = self._get_server_name(search_term)
        return await super().aio_run(str(search_term), server)



import asyncwhois


def main():
    domain = "google.com"

    # show the entire query-chain answer
    result = asyncwhois.whois_domain(domain, authoritative_only=False)  # default
    print(result.query_output)

    # only show the "authoritative" answer
    result = asyncwhois.whois_domain(domain, authoritative_only=True)
    print(result.query_output)

    return


if __name__ == "__main__":
    main()
